package com.uoc.cclavero.freeauthenticator.activities;

import android.Manifest;
import android.arch.lifecycle.Observer;
import android.arch.lifecycle.ViewModelProviders;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.pm.PackageManager;
import android.os.Bundle;
import android.support.annotation.NonNull;
import android.support.annotation.Nullable;
import android.support.design.widget.FloatingActionButton;
import android.support.v4.app.ActivityCompat;
import android.support.v4.content.ContextCompat;
import android.support.v7.app.AlertDialog;
import android.support.v7.app.AppCompatActivity;
import android.support.v7.widget.LinearLayoutManager;
import android.support.v7.widget.RecyclerView;
import android.support.v7.widget.Toolbar;
import android.util.Base64;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.widget.EditText;
import android.widget.Toast;
import android.os.Handler;

import com.google.gson.Gson;
import com.google.gson.reflect.TypeToken;
import com.uoc.cclavero.freeauthenticator.R;
import com.uoc.cclavero.freeauthenticator.adapters.KeyListAdapter;
import com.uoc.cclavero.freeauthenticator.data.KeyViewModel;
import com.uoc.cclavero.freeauthenticator.dialogs.AddKeyDialog;
import com.uoc.cclavero.freeauthenticator.entities.Key;
import com.uoc.cclavero.freeauthenticator.helpers.Base32String;
import com.uoc.cclavero.freeauthenticator.helpers.EncryptionHelper;
import com.uoc.cclavero.freeauthenticator.helpers.ExternalStorageHelper;
import com.uoc.cclavero.freeauthenticator.helpers.OneTimePasswordAlgorithm;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;

import javax.crypto.SecretKey;

public class MainActivity extends AppCompatActivity implements AddKeyDialog.AddKeyDialogListener {

    private static final int ADD_KEY_KEYBOARD_ACTIVITY_REQUEST_CODE = 1;
    private static final int ADD_KEY_QR_ACTIVITY_REQUEST_CODE = 2;
    private static final int CREATE_BACKUP_ACTIVITY_REQUEST_CODE = 3;
    private static final int RESTORE_BACKUP_ACTIVITY_REQUEST_CODE = 4;
    private static final int PERMISSIONS_REQUEST_CAMERA = 1;
    private static final int PERMISSIONS_REQUEST_WRITE_EXTERNAL_STORAGE_CREATE_BACKUP = 2;
    private static final int PERMISSIONS_REQUEST_WRITE_EXTERNAL_STORAGE_RESTORE_BACKUP = 3;
    public static final int TOKEN_DIGITS = 6;
    public static final int DEFAULT_DURATION = 30;
    private static final int TRUNCATION_OFFSET = 16;
    public static final String AES_KEY_FILE = "aes.key";

    private KeyViewModel keyViewModel;
    private Handler handler;
    private Runnable handlerTask;
    private Context context;
    private KeyListAdapter adapter;
    private boolean firstRun = true;


    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);
        Toolbar toolbar = findViewById(R.id.toolbar);
        setSupportActionBar(toolbar);
        context = this;

        FloatingActionButton fab = findViewById(R.id.fab);
        fab.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                AddKeyDialog addKey = new AddKeyDialog();
                addKey.show(getSupportFragmentManager(), "addKeyDialog");
            }
        });

        RecyclerView recyclerView = findViewById(R.id.recyclerview);
        recyclerView.setLayoutManager(new LinearLayoutManager(this));
        recyclerView.setHasFixedSize(true);

        adapter = new KeyListAdapter(this);
        recyclerView.setAdapter(adapter);

        keyViewModel = ViewModelProviders.of(this).get(KeyViewModel.class);

        keyViewModel.getAllKeys().observe(this, new Observer<List<Key>>() {
            @Override
            public void onChanged(@Nullable List<Key> keys) {
                adapter.setKeys(keys);
            }
        });

        adapter.setOnItemClickListener(new KeyListAdapter.OnItemMenuClickListener() {
            @Override
            public void onItemMenuClick(final Key key, boolean delete) {
                AlertDialog.Builder alert = new AlertDialog.Builder(context);
                if(!delete){
                    alert.setTitle(R.string.title_dialog_rename_alert);
                    alert.setMessage(" ");

                    final EditText input = new EditText(context);
                    input.setText(key.getDescription());
                    input.setSelection(key.getDescription().length());

                    alert.setView(input);


                    alert.setPositiveButton(R.string.button_save, new DialogInterface.OnClickListener() {
                        public void onClick(DialogInterface dialog, int whichButton) {
                            key.setDescription(input.getEditableText().toString());
                            keyViewModel.update(key);
                        }
                    });

                    alert.setNegativeButton(R.string.button_cancel, new DialogInterface.OnClickListener() {
                        public void onClick(DialogInterface dialog, int whichButton) {
                            dialog.cancel();
                        }
                    });
                } else {
                    alert.setTitle(R.string.title_dialog_delete_alert);
                    alert.setMessage(R.string.message_delete_alert);

                    alert.setPositiveButton(R.string.button_delete, new DialogInterface.OnClickListener() {
                        public void onClick(DialogInterface dialog, int whichButton) {
                            keyViewModel.delete(key);
                        }
                    });

                    alert.setNegativeButton(R.string.button_cancel, new DialogInterface.OnClickListener() {
                        public void onClick(DialogInterface dialog, int whichButton) {
                            dialog.cancel();
                        }
                    });
                }

                alert.show();
            }
        });

        /*
         * Main Loop
         * Here is where tokens are generated outside of UI Thread
         */

        handler = new Handler();
        handlerTask = new Runnable()
        {
            @Override
            public void run() {
                handler.postDelayed(this, 1000);

                Key currentKey;
                String token;
                SecretKey aesKey;

                long currentTime = System.currentTimeMillis();
                int progress = 30 - (int) ((currentTime / 1000) % 30);

                try {
                    aesKey = EncryptionHelper.getOrGenerateAESKey(context, new File(context.getFilesDir() + "/" + AES_KEY_FILE));
                } catch (Exception e) {
                    Toast.makeText(context,R.string.message_error_reading_private_key, Toast.LENGTH_LONG).show();
                    return;
                }
                for(int i =0;i < adapter.getItemCount(); i++){
                    currentKey = adapter.getKeyAt(i);
                    adapter.setCountDown(progress);
                    if(firstRun || (!firstRun && progress == 30))
                    {
                        token = generateToken(currentKey, currentTime, aesKey);
                        currentKey.setToken(token);
                        keyViewModel.update(currentKey);
                    }
                }
                adapter.notifyDataSetChanged();
                firstRun = false;
            }
        };
    }

    private String generateToken(Key key, Long currentTime, SecretKey aesKey) {
        try{
            //Moving Factor = Current Time in seconds (epoch) / duration time of the code
            Long movingFactor = currentTime/1000/DEFAULT_DURATION;
            byte[] decryptedKey = EncryptionHelper.decrypt(aesKey,key.getPrivateKey());

            return OneTimePasswordAlgorithm.generateOTP(decryptedKey, movingFactor, TOKEN_DIGITS,false,TRUNCATION_OFFSET);
        }catch (Exception e){
            Toast.makeText(context,"Failed to generate Token.", Toast.LENGTH_SHORT).show();
            return "";
        }
    }

    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        getMenuInflater().inflate(R.menu.menu_main, menu);
        return true;
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        int id = item.getItemId();

        switch (id){
            case R.id.action_create_backup:
                if (ContextCompat.checkSelfPermission(this, Manifest.permission.WRITE_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED) {
                    ActivityCompat.requestPermissions(this, new String[]{Manifest.permission.WRITE_EXTERNAL_STORAGE}, PERMISSIONS_REQUEST_WRITE_EXTERNAL_STORAGE_CREATE_BACKUP);
                }
                else {
                    openCreateBackupActivity();
                }
                break;
            case R.id.action_restore_backup:
                if (ContextCompat.checkSelfPermission(this, Manifest.permission.WRITE_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED) {
                    ActivityCompat.requestPermissions(this, new String[]{Manifest.permission.WRITE_EXTERNAL_STORAGE}, PERMISSIONS_REQUEST_WRITE_EXTERNAL_STORAGE_RESTORE_BACKUP);
                }
                else {
                    openRestoreBackupActivity();
                }
                break;
            default:
                return false;
        }

        return false;
    }

    @Override
    public void onResume() {
        super.onResume();

        handler.post(handlerTask);
    }

    @Override
    public void onPause() {
        super.onPause();

        handler.removeCallbacks(handlerTask);
    }

    @Override
    public void onButtonClicked(String text) {
        switch (text){
            case "qr_scanner":
                if (ContextCompat.checkSelfPermission(this, Manifest.permission.CAMERA) != PackageManager.PERMISSION_GRANTED) {
                    ActivityCompat.requestPermissions(this, new String[]{Manifest.permission.CAMERA}, PERMISSIONS_REQUEST_CAMERA);
                }
                else {
                    openQRCameraActivity();
                }
                break;
            case "keyboard_input":
                Intent startChildActivityIntent = new Intent(MainActivity.this, AddKeyKeyboardActivity.class);
                startActivityForResult(startChildActivityIntent, ADD_KEY_KEYBOARD_ACTIVITY_REQUEST_CODE);
                break;
            default:
                break;
        }
    }

    private void openQRCameraActivity() {
        Intent startChildActivityIntent = new Intent(MainActivity.this, AddKeyQRActivity.class);
        startActivityForResult(startChildActivityIntent, ADD_KEY_QR_ACTIVITY_REQUEST_CODE);
    }

    private void openCreateBackupActivity() {
        Intent startChildActivityIntent = new Intent(MainActivity.this, CreateBackupActivity.class);
        startActivityForResult(startChildActivityIntent, CREATE_BACKUP_ACTIVITY_REQUEST_CODE);
    }

    private void openRestoreBackupActivity() {
        Intent startChildActivityIntent = new Intent(MainActivity.this, RestoreBackupActivity.class);
        startActivityForResult(startChildActivityIntent, RESTORE_BACKUP_ACTIVITY_REQUEST_CODE);
    }


    public void onActivityResult(int requestCode, int resultCode, Intent data) {
        super.onActivityResult(requestCode, resultCode, data);

        if(resultCode == RESULT_OK){
            try{
                long currentTime = System.currentTimeMillis();
                switch (requestCode) {
                    case ADD_KEY_KEYBOARD_ACTIVITY_REQUEST_CODE:
                    case ADD_KEY_QR_ACTIVITY_REQUEST_CODE:
                        SecretKey aesKey = EncryptionHelper.getOrGenerateAESKey(context, new File(context.getFilesDir() + "/" + AES_KEY_FILE));
                        String keyPrivate = data.getStringExtra(AddKeyQRActivity.KEY);
                        byte[] encryptedKey = EncryptionHelper.encrypt(aesKey,Base32String.decode(keyPrivate));
                        Key key = new Key(data.getStringExtra(AddKeyQRActivity.DESCRIPTION), Base64.encodeToString(encryptedKey, Base64.DEFAULT));

                        key.setToken(generateToken(key,currentTime, aesKey));
                        keyViewModel.insert(key);
                        break;
                    case CREATE_BACKUP_ACTIVITY_REQUEST_CODE:
                        storeBackup(data.getStringExtra(CreateBackupActivity.DESTINATION_STORAGE), data.getStringExtra(CreateBackupActivity.ENCRYPTION_KEY));
                        break;
                    case RESTORE_BACKUP_ACTIVITY_REQUEST_CODE:
                        restoreBackup(data.getStringExtra(RestoreBackupActivity.ENCRYPTION_KEY), data.getStringExtra(RestoreBackupActivity.FILE_CONTENT));
                        break;
                }
            }catch (Exception e) {
                Toast.makeText(this, R.string.message_error_invalid_private_key, Toast.LENGTH_LONG).show();
            }

        }
    }

    @Override
    public void onRequestPermissionsResult(int requestCode, @NonNull String permissions[], @NonNull int[] grantResults) {
        switch (requestCode) {
            case PERMISSIONS_REQUEST_CAMERA:
                if (grantResults.length > 0 && grantResults[0] == PackageManager.PERMISSION_GRANTED) {
                    Intent startChildActivityIntent = new Intent(MainActivity.this, AddKeyQRActivity.class);
                    startActivityForResult(startChildActivityIntent, ADD_KEY_QR_ACTIVITY_REQUEST_CODE);
                } else {
                    Toast.makeText(this, R.string.message_error_grant_camera_permission, Toast.LENGTH_LONG).show();
                }
                return;
            case PERMISSIONS_REQUEST_WRITE_EXTERNAL_STORAGE_CREATE_BACKUP:
                if (grantResults.length > 0 && grantResults[0] == PackageManager.PERMISSION_GRANTED) {
                    Intent startChildActivityIntent = new Intent(MainActivity.this, CreateBackupActivity.class);
                    startActivity(startChildActivityIntent);
                } else {
                    Toast.makeText(this, R.string.message_error_grant_write_external_storage_permission, Toast.LENGTH_LONG).show();
                }
                return;
            case PERMISSIONS_REQUEST_WRITE_EXTERNAL_STORAGE_RESTORE_BACKUP:
                if (grantResults.length > 0 && grantResults[0] == PackageManager.PERMISSION_GRANTED) {
                    Intent startChildActivityIntent = new Intent(MainActivity.this, CreateBackupActivity.class);
                    startActivity(startChildActivityIntent);
                } else {
                    Toast.makeText(this, R.string.message_error_grant_write_external_storage_permission, Toast.LENGTH_LONG).show();
                }
                return;
        }
    }

    private void storeBackup(String destinationStorage, String encryptionKey) {
        String encryptedJsonKeys = getDatabaseAsEncryptedJsonArrayString(encryptionKey);
        if(!encryptedJsonKeys.equals("")) {
            SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd");
            String currentDate = sdf.format(new Date());
            String filename = "FreeAuthenticator_"+currentDate+".txt";

            switch(destinationStorage){
                case "local":
                    if(ExternalStorageHelper.isExternalStorageWritable())
                    {
                        File backupFile = ExternalStorageHelper.storeFileInPublicStorageFolder(filename);

                        if(backupFile != null && !ExternalStorageHelper.writeStringToFile(backupFile, encryptedJsonKeys))
                        {
                            Toast.makeText(this, R.string.message_error_store_backup, Toast.LENGTH_LONG).show();
                        }

                        Toast.makeText(this, R.string.message_store_backup_success, Toast.LENGTH_LONG).show();
                    }
                    else {
                        Toast.makeText(this, R.string.message_error_storage_not_writable, Toast.LENGTH_LONG).show();
                    }
                    break;
                case "google_drive":
                    Intent intent = new Intent(getBaseContext(), GoogleDriveCreateFileActivity.class);
                    intent.putExtra(GoogleDriveCreateFileActivity.FILENAME, filename);
                    intent.putExtra(GoogleDriveCreateFileActivity.FILE_CONTENT, encryptedJsonKeys);
                    startActivity(intent);
                    break;
            }
        }
    }

    private String getDatabaseAsEncryptedJsonArrayString(String password){
        StringBuilder jsonArray = new StringBuilder();
        Key currentKey;
        SecretKey aesKey;

        jsonArray.append("[");
        try {
            aesKey = EncryptionHelper.getOrGenerateAESKey(context, new File(context.getFilesDir() + "/" + AES_KEY_FILE));
            for(int i = 0; i < adapter.getItemCount(); i++){
                currentKey = adapter.getKeyAt(i);
                if(i > 0)
                    jsonArray.append(",");
                String privateKey = Base32String.encode(EncryptionHelper.decrypt(aesKey,currentKey.getPrivateKey()));
                jsonArray.append("{\"privateKey\":\"").append(privateKey).append("\",")
                        .append("\"description\":\"").append(currentKey.getDescription())
                        .append("\"}");
            }

            jsonArray.append("]");
            return EncryptionHelper.encrypt(password, jsonArray.toString());
        } catch (Exception e) {
            Toast.makeText(context,R.string.message_error_extracting_backup_data, Toast.LENGTH_LONG).show();
            return "";
        }
    }

    private void restoreBackup(String password, String encryptedFileContent) {
        List<Key> decryptedKeys = getKeysListFromEncryptedFile(password, encryptedFileContent);
        Key currentKey;
        SecretKey aesKey;
        Long currentTime = System.currentTimeMillis();

        try {
            if(decryptedKeys != null) {
                aesKey = EncryptionHelper.getOrGenerateAESKey(context, new File(context.getFilesDir() + "/" + AES_KEY_FILE));
                for (int i = 0; i < decryptedKeys.size(); i++) {
                    currentKey = decryptedKeys.get(i);
                    byte[] encryptedKey = EncryptionHelper.encrypt(aesKey,Base32String.decode(currentKey.getPrivateKey()));
                    currentKey.setPrivateKey(Base64.encodeToString(encryptedKey, Base64.DEFAULT));
                    currentKey.setToken(generateToken(currentKey,currentTime, aesKey));
                    keyViewModel.insert(currentKey);
                }
            }
        } catch (Exception e) {
            Toast.makeText(context,R.string.message_error_reading_private_key, Toast.LENGTH_LONG).show();
        }
    }



    private List<Key> getKeysListFromEncryptedFile(String password, String encryptedJsonContent) {
        try {
            String decryptedJsonString = EncryptionHelper.decrypt(password, encryptedJsonContent);

            Gson gson = new Gson();
            return gson.fromJson(decryptedJsonString, new TypeToken<List<Key>>(){}.getType());
        } catch(Exception e) {
            Toast.makeText(context,R.string.message_error_restoring_backup, Toast.LENGTH_LONG).show();
            return null;
        }
    }

}
